package net.wanji.opt.service.impl;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.date.DateUnit;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.util.ObjectUtil;
import net.wanji.common.dto.CrossIdDTO;
import net.wanji.common.enums.CrossStatusEnum;
import net.wanji.common.enums.EventAlarmEnum;
import net.wanji.common.enums.EventAlarmSourceEnum;
import net.wanji.common.utils.tool.CrossUtil;
import net.wanji.databus.dao.entity.*;
import net.wanji.databus.dao.mapper.*;
import net.wanji.databus.po.*;
import net.wanji.databus.vo.AbnormalCrossListVO;
import net.wanji.opt.bo.GreenwaveIdAndTimeStampBO;
import net.wanji.opt.bo.GreenwaveIdBO;
import net.wanji.opt.bo.SaveGreenwaveStrategyBO;
import net.wanji.opt.dao.mapper.CrossSchemeOptLogMapper;
import net.wanji.opt.dao.mapper.strategy.SceneStrategyIdeaMapper;
import net.wanji.opt.dao.mapper.strategy.StrategyMapper;
import net.wanji.opt.dao.mapper.trend.EventAlarmMapper;
import net.wanji.opt.dao.mapper.trend.GreenwaveInfoMapper;
import net.wanji.opt.dto.trend.AbnormalCrossListDTO;
import net.wanji.opt.dto.trend.GreenwaveListDTO;
import net.wanji.opt.po.base.CrossSchemeOptLogPO;
import net.wanji.opt.po.strategy.StrategyPO;
import net.wanji.opt.po.trend.EventAlarmPO;
import net.wanji.opt.service.TrendService;
import net.wanji.opt.vo.*;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @author Kent HAN
 * @date 2023/2/8 10:31
 */
@Service
public class TrendServiceImpl implements TrendService {
    private final GreenwaveInfoMapper greenwaveInfoMapper;
    private final BaseCrossInfoMapper baseCrossInfoMapper;
    private final CrossDataRealtimeMapper crossDataRealtimeMapper;
    private final CrossTurnDataRealtimeMapper crossTurnDataRealtimeMapper;
    private final CrossDirDataHistMapper crossDirDataHistMapper;
    private final EventAlarmMapper eventAlarmMapper;
    private final RidInfoMapper ridInfoMapper;
    private final BaseCrossDirInfoMapper baseCrossDirInfoMapper;
    private final CrossSchemeOptLogMapper crossSchemeOptLogMapper;
    private final GreenwaveCrossMapper greenwaveCrossMapper;
    private final CrossDirDataRealtimeMapper crossDirDataRealtimeMapper;
    private final GreenwaveHistMapper greenwaveHistMapper;
    private final GreenwaveRealtimeMapper greenwaveRealtimeMapper;
    private final StrategyMapper strategyMapper;
    private final BaseCrossSchedulesMapper baseCrossSchedulesMapper;
    private final CrossSchedulesPlanMapper crossSchedulesPlanMapper;
    private final BaseCrossSectionMapper baseCrossSectionMapper;
    private final BaseCrossSchemeMapper baseCrossSchemeMapper;
    private final BaseCrossPhaseMapper baseCrossPhaseMapper;

    SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");

    public TrendServiceImpl(GreenwaveInfoMapper greenwaveInfoMapper, BaseCrossInfoMapper baseCrossInfoMapper,
                            CrossDataRealtimeMapper crossDataRealtimeMapper,
                            CrossTurnDataRealtimeMapper crossTurnDataRealtimeMapper,
                            CrossDirDataHistMapper crossDirDataHistMapper, EventAlarmMapper eventAlarmMapper,
                            RidInfoMapper ridInfoMapper, BaseCrossDirInfoMapper baseCrossDirInfoMapper,
                            CrossSchemeOptLogMapper crossSchemeOptLogMapper, GreenwaveCrossMapper greenwaveCrossMapper, CrossDirDataRealtimeMapper crossDirDataRealtimeMapper, GreenwaveHistMapper greenwaveHistMapper, GreenwaveRealtimeMapper greenwaveRealtimeMapper, SceneStrategyIdeaMapper strategyIdeaMapper, StrategyMapper strategyMapper, BaseCrossSchedulesMapper baseCrossSchedulesMapper, CrossSchedulesPlanMapper crossSchedulesPlanMapper, BaseCrossSectionMapper baseCrossSectionMapper, BaseCrossSchemeMapper crossSchemeMapper, BaseCrossSchemeMapper baseCrossSchemeMapper, BaseCrossPhaseMapper baseCrossPhaseMapper) {
        this.greenwaveInfoMapper = greenwaveInfoMapper;
        this.baseCrossInfoMapper = baseCrossInfoMapper;
        this.crossDataRealtimeMapper = crossDataRealtimeMapper;
        this.crossTurnDataRealtimeMapper = crossTurnDataRealtimeMapper;
        this.crossDirDataHistMapper = crossDirDataHistMapper;
        this.eventAlarmMapper = eventAlarmMapper;
        this.ridInfoMapper = ridInfoMapper;
        this.baseCrossDirInfoMapper = baseCrossDirInfoMapper;
        this.crossSchemeOptLogMapper = crossSchemeOptLogMapper;
        this.greenwaveCrossMapper = greenwaveCrossMapper;
        this.crossDirDataRealtimeMapper = crossDirDataRealtimeMapper;
        this.greenwaveHistMapper = greenwaveHistMapper;
        this.greenwaveRealtimeMapper = greenwaveRealtimeMapper;
        this.strategyMapper = strategyMapper;
        this.baseCrossSchedulesMapper = baseCrossSchedulesMapper;
        this.crossSchedulesPlanMapper = crossSchedulesPlanMapper;
        this.baseCrossSectionMapper = baseCrossSectionMapper;
        this.baseCrossSchemeMapper = baseCrossSchemeMapper;
        this.baseCrossPhaseMapper = baseCrossPhaseMapper;
    }

    @Override
    public List<GreenwaveListVO> greenwaveList(GreenwaveListDTO greenwaveListDTO) {
        Integer status = greenwaveListDTO.getStatus();
        String name = greenwaveListDTO.getName();
        Integer type = greenwaveListDTO.getType();
        List<GreenwaveListVO> greenwaveListVOList = greenwaveInfoMapper.listGreenwave(status, name, type);
        // 3、4都算拥堵
        if (status != null && status == 3) {
            List<GreenwaveListVO> extraList = greenwaveInfoMapper.listGreenwave(4, name, type);
            greenwaveListVOList.addAll(extraList);
        }
        changeStatus(greenwaveListVOList);
        for (GreenwaveListVO greenwaveListVO : greenwaveListVOList) {
            // 计算持续时间
            buildDuration(greenwaveListVO);
            // 构造路口列表
            Integer greenwaveId = greenwaveListVO.getId();
            List<GreenwaveListVO.CrossListElement> crossList = buildCrossList(greenwaveId);
            greenwaveListVO.setCrossList(crossList);
        }
        return greenwaveListVOList;
    }

    @Override
    public GreenwaveStats buildGreenwaveStats(List<GreenwaveListVO> greenwaveListVOList) {
        GreenwaveStats greenwaveStats = new GreenwaveStats();
        Integer smooth = greenwaveStats.getSmooth();
        Integer slow = greenwaveStats.getSlow();
        Integer congestion = greenwaveStats.getCongestion();
        for (GreenwaveListVO greenwaveListVO : greenwaveListVOList) {
            Integer realtimeStatus = greenwaveListVO.getRealtimeStatus();
            if (realtimeStatus == 1) {
                smooth += 1;
            } else if (realtimeStatus == 2) {
                slow += 1;
            } else if (realtimeStatus == 3) {
                congestion += 1;
            }
        }
        greenwaveStats.setSmooth(smooth);
        greenwaveStats.setSlow(slow);
        greenwaveStats.setCongestion(congestion);
        return greenwaveStats;
    }

    @Override
    public AbnormalCrossVO abnormalCrossList(AbnormalCrossListDTO abnormalCrossListDTO) {
        AbnormalCrossVO abnormalCrossVO = new AbnormalCrossVO();
        // 构建问题路口列表
        Integer status = abnormalCrossListDTO.getStatus();
        String name = abnormalCrossListDTO.getName();
        Integer type = abnormalCrossListDTO.getType();
        List<AbnormalCrossListVO> abnormalCrossListVOList =
                crossDataRealtimeMapper.selectAbnormalCross(status, name, type);
        // 添加优化状态
        for (AbnormalCrossListVO abnormalCrossListVO : abnormalCrossListVOList) {
            abnormalCrossListVO.setOptStatus(2); // 2未优化
            String crossId = abnormalCrossListVO.getId();
            // 查询优化日志
            List<CrossSchemeOptLogPO> crossSchemeOptLogPOS = crossSchemeOptLogMapper.selectMaxByCrossId(crossId);
            if (CollectionUtil.isNotEmpty(crossSchemeOptLogPOS)) {
                CrossSchemeOptLogPO crossSchemeOptLogPO = crossSchemeOptLogPOS.get(0);
                Date endTime = crossSchemeOptLogPO.getEndTime();
                if (ObjectUtil.isEmpty(endTime)) {
                    abnormalCrossListVO.setOptStatus(1); // 1优化中
                }
            }
        }
        // 坐标格式转换
        for (AbnormalCrossListVO abnormalCrossListVO : abnormalCrossListVOList) {
            String locationStr = abnormalCrossListVO.getLocationStr();
            double[] lonLat = CrossUtil.getLonLat(locationStr);
            List<Double> location = new ArrayList<>();
            location.add(lonLat[0]);
            location.add(lonLat[1]);
            abnormalCrossListVO.setLocation(location);
            // 关联路口信息
            Integer realtimeStatus = abnormalCrossListVO.getRealtimeStatus();
            String endCrossId = abnormalCrossListVO.getId();
            if (Objects.equals(realtimeStatus, CrossStatusEnum.SPILLOVER.getCode())) {
                // 溢出只返回溢出方向的关联路口
                String spilloverDirs = abnormalCrossListVO.getSpilloverDirs();
                String[] split = spilloverDirs.split(",");
                List<Integer> spilloverDirList = new ArrayList<>();
                for (String s : split) {
                    spilloverDirList.add(Integer.parseInt(s));
                }
                buildRidData(abnormalCrossListVO, endCrossId, spilloverDirList);
            } else if (Objects.equals(realtimeStatus, CrossStatusEnum.CONGESTION.getCode())){
                // 拥堵返回所有方向关联路口
                List<Integer> inDirs = baseCrossDirInfoMapper.selectInDirsByCrossId(endCrossId);
                buildRidData(abnormalCrossListVO, endCrossId, inDirs);
            }
        }
        CrossComparator crossComparator = new CrossComparator();
        abnormalCrossListVOList.sort(crossComparator);
        abnormalCrossVO.setAbnormalCrossList(abnormalCrossListVOList);
        // 构造统计信息
        List<AbnormalCrossListVO> listForStats =
                crossDataRealtimeMapper.selectAbnormalCross(null, null, null);
        buildAbnormalCrossStats(abnormalCrossVO, listForStats);

        return abnormalCrossVO;
    }

    private static class CrossComparator implements Comparator<AbnormalCrossListVO> {
        @Override
        public int compare(AbnormalCrossListVO o1, AbnormalCrossListVO o2) {
            // 先按路口状态排序
            Integer o1status = o1.getRealtimeStatus();
            Integer o2status = o2.getRealtimeStatus();
            int flag = o2status - o1status;
            if (flag == 0) {
                // 再按拥堵时长排序
                Long o1duration = o1.getDuration();
                Long o2Duration = o2.getDuration();
                long res = o2Duration - o1duration;
                return (int) res;
            }
            return flag;
        }
    }

    private void buildRidData(AbnormalCrossListVO abnormalCrossListVO, String endCrossId,
                              List<Integer> spilloverDirList) {
        List<List<Double>> ridLocationList = new ArrayList<>();
        List<String> ridWktList = new ArrayList<>();
        for (Integer spilloverDir : spilloverDirList) {
            RidInfoEntity ridInfo = ridInfoMapper.selectByEndInDir(endCrossId, spilloverDir);
            if (ridInfo != null) {
                // 关联路口坐标
                String startCrossId = ridInfo.getStartCrossId();
                TBaseCrossInfo crossInfoPO = baseCrossInfoMapper.selectByPrimaryKey(startCrossId);
                if (crossInfoPO != null) {
                    String startLocationStr = crossInfoPO.getLocation();
                    double[] startLonLat = CrossUtil.getLonLat(startLocationStr);
                    List<Double> startLocation = new ArrayList<>();
                    startLocation.add(startLonLat[0]);
                    startLocation.add(startLonLat[1]);
                    ridLocationList.add(startLocation);
                }
                // 关联路口wkt
                String wkt = ridInfo.getWkt();
                ridWktList.add(wkt);
            }
        }
        abnormalCrossListVO.setRidLocationList(ridLocationList);
        abnormalCrossListVO.setRidWktList(ridWktList);
    }

    @Override
    public AbnormalCrossDetailVO abnormalCrossDetail(CrossIdDTO crossIdDTO) {
        String crossId = crossIdDTO.getCrossId();
        AbnormalCrossDetailVO abnormalCrossDetailVO = new AbnormalCrossDetailVO();

        // 构造路口实时数据
        CrossDataRealtimePO crossDataRealtimePO = crossDataRealtimeMapper.selectByCrossId(crossId);
        abnormalCrossDetailVO.setCrossStatus(crossDataRealtimePO.getStatus());
        abnormalCrossDetailVO.setCrossQueueLength(crossDataRealtimePO.getQueueLength());
        abnormalCrossDetailVO.setCrossFlowRate((double)crossDataRealtimePO.getFlow());
        abnormalCrossDetailVO.setCrossStopTimes(crossDataRealtimePO.getStopTimes());
        abnormalCrossDetailVO.setCongestionIndex(crossDataRealtimePO.getCongestionIndex());
        // 构造转向实时数据
        List<AbnormalCrossDetailVO.TurnDataElement> turnData = buildTurnData(crossId);
        abnormalCrossDetailVO.setTurnData(turnData);
        // 构造方向历史数据
        List<AbnormalCrossDetailVO.DirDataElement> dirData = buildDirData(crossId);
        abnormalCrossDetailVO.setDirData(dirData);

        return abnormalCrossDetailVO;
    }

    @Override
    public List<EventAlarmVO> eventAlarm(Date todayTime) throws ParseException {
        List<EventAlarmPO> eventAlarmPOList = eventAlarmMapper.selectByTodayTime(todayTime);
        List<EventAlarmVO> eventAlarmOutVOList = new ArrayList<>();
        for (EventAlarmPO eventAlarmPO : eventAlarmPOList) {
            EventAlarmVO eventAlarmOutVO = new EventAlarmVO();
            eventAlarmOutVO.setEventDesc(eventAlarmPO.getEventDesc());
            // 事件类型
            Integer typeCode = eventAlarmPO.getType();
            String msgByCode = EventAlarmEnum.getMsgByCode(typeCode);
            eventAlarmOutVO.setType(msgByCode);

            Date endTime = eventAlarmPO.getEndTime();
            if (typeCode == 1 || typeCode == 2) {
                // 发生时间
                Date startTime = eventAlarmPO.getStartTime();
                String startTimeStr = sdf.format(startTime);
                String substring = startTimeStr.substring(11, 16);
                eventAlarmOutVO.setStartTime(substring);
                // 时长
                if (endTime == null) {
                    Date nowTime = DateUtil.date();
                    long between = DateUtil.between(startTime, nowTime, DateUnit.MINUTE);
                    int betweenInt = (int) between;
                    eventAlarmOutVO.setDuration(betweenInt);
                } else {
                    long between = DateUtil.between(startTime, endTime, DateUnit.MINUTE);
                    int betweenInt = (int) between;
                    eventAlarmOutVO.setDuration(betweenInt);
                }
            }
            if (typeCode == 3 || typeCode == 4) {
                // 时间范围
                Date startTime = eventAlarmPO.getStartTime();
                String startTimeStr = sdf.format(startTime);
                String startTimeSubStr = startTimeStr.substring(0, 16); // 截取到分钟
                String endTimeStr = sdf.format(endTime);
                String endTimeSubStr = endTimeStr.substring(0, 16); // 截取到分钟
                eventAlarmOutVO.setTimeSpan(startTimeSubStr + "~" + endTimeSubStr);
            }

            Integer sourceCode = eventAlarmPO.getSource();
            String sourceStr = EventAlarmSourceEnum.getMsgByCode(sourceCode);
            eventAlarmOutVO.setSource(sourceStr);
            eventAlarmOutVO.setCoordinate(eventAlarmPO.getWkt());

            eventAlarmOutVOList.add(eventAlarmOutVO);

        }
        return eventAlarmOutVOList;
    }

    @Override
    public List<GreenWaveCrossMonitorVO> greenwaveCrossMonitor(GreenwaveIdBO greenwaveIdBO) {
        Integer greenwaveId = greenwaveIdBO.getGreenwaveId();
        List<GreenwaveCrossPO> greenwaveCrossPOList = greenwaveCrossMapper.selectByGreenwaveId(greenwaveId);
        List<GreenWaveCrossMonitorVO> res = buildRes(greenwaveCrossPOList);
        return res;
    }

    @Override
    public List<GreenwaveRunMonitorVO> greenwaveRunMonitor(GreenwaveIdBO greenwaveIdBO) {
        List<GreenwaveRunMonitorVO> res = new ArrayList<>();
        Date nowTime = new Date();
        // todo 测试用，固定当前时间
        nowTime = DateUtil.parse("2023-06-09 15:00:00","yyyy-MM-dd HH:mm:ss");
        List<GreenwaveHistPOExt> extList = greenwaveHistMapper.selectRunMonitor(nowTime);
        for (GreenwaveHistPOExt greenwaveHistPOExt : extList) {
            GreenwaveRunMonitorVO greenWaveRunMonitorVO = new GreenwaveRunMonitorVO();
            String hour = greenwaveHistPOExt.getHour();
            String hourSub = hour.substring(11, 16); // 截取小时
            greenWaveRunMonitorVO.setHour(hourSub);
            greenWaveRunMonitorVO.setAvgSpeed(greenwaveHistPOExt.getAvgSpeed());
            greenWaveRunMonitorVO.setCongestionIndex(greenwaveHistPOExt.getCongestionIndex());
            res.add(greenWaveRunMonitorVO);
        }
        return res;
    }

    @Override
    public GreenwaveDetailVO greenwaveDetail(GreenwaveIdBO greenwaveIdBO) {
        Integer greenwaveId = greenwaveIdBO.getGreenwaveId();
        GreenwaveDetailVO res = new GreenwaveDetailVO();
        GreenwaveRealtimePO greenwaveRealtimePO = greenwaveRealtimeMapper.selectById(greenwaveId);
        GreenwaveInfoPO greenwaveInfoPO = greenwaveInfoMapper.selectById(greenwaveId);
        res.setName(greenwaveInfoPO.getName());
        res.setStatus(greenwaveInfoPO.getStatus());

        double uncoordinatePhaseQueue = greenwaveRealtimePO.getUncoordinatePhaseQueue();
        res.setUncoordinatePhaseQueue((int)uncoordinatePhaseQueue);
        double noparkPassRate = greenwaveRealtimePO.getNoparkPassRate();
        res.setNoparkPassRate((int) (noparkPassRate * 100));
        res.setTrvalTime(greenwaveRealtimePO.getTrvalTime());
        // 获取策略信息
        List<String> strategyList = strategyMapper.selectAllNames();
        res.setStrategyList(strategyList);
        List<StrategyPO> strategyPOList = strategyMapper.selectByGreenwaveId(greenwaveId);
        List<String> collect = strategyPOList.stream()
                .map(StrategyPO::getStrategyName)
                .collect(Collectors.toList());
        res.setStrategySelected(collect);

        List<GreenwaveDetailVO.GreenwaveCross> greenwaveCrossList = buildGreenwaveCrossList(greenwaveId);
        res.setGreenwaveCrossList(greenwaveCrossList);

        return res;
    }

    private List<GreenwaveDetailVO.GreenwaveCross> buildGreenwaveCrossList(Integer greenwaveId) {
        List<GreenwaveDetailVO.GreenwaveCross> res = new ArrayList<>();
        List<GreenwaveCrossPO> greenwaveCrossList = greenwaveCrossMapper.selectByGreenwaveId(greenwaveId);
        int size = greenwaveCrossList.size();
        for (int i = 0; i < size; i++) {
            GreenwaveCrossPO greenwaveCrossPO = greenwaveCrossList.get(i);
            GreenwaveDetailVO.GreenwaveCross greenwaveCross = new GreenwaveDetailVO.GreenwaveCross();
            String crossId = greenwaveCrossPO.getCrossId();
            greenwaveCross.setCrossId(crossId);
            BaseCrossInfoPO baseCrossInfoPO = baseCrossInfoMapper.selectById(crossId);
            greenwaveCross.setCrossName(baseCrossInfoPO.getName());
            // 根据路段获取距离
            Integer distanceToNextCross = null;
            if (i < size - 1) { // 不是最后一个路口
                distanceToNextCross = calcDistance(greenwaveCrossList, i, crossId);
            }
            greenwaveCross.setDistanceToNextCross(distanceToNextCross);
            // 查找方向平均速度
            Double speed = calcSpeed(greenwaveCrossList, i, crossId);
            greenwaveCross.setSpeed(speed);

            // 根据当前时间获取方案信息
            Date todayDate = new Date();
            // 获取当前执行的调度
            BaseCrossSchedulesPO baseCrossSchedulesPO = baseCrossSchedulesMapper.selectExecByCrossId(crossId);
            Integer schedulesId = baseCrossSchedulesPO.getId();
            // 获取日计划
            Integer planId = null;
            // 判断是否是特殊日期
            CrossSchedulesPlanPO crossSchedulesPlanPO =
                    crossSchedulesPlanMapper.selectBySpecialDate(crossId, schedulesId, todayDate);
            if (ObjectUtil.isNotEmpty(crossSchedulesPlanPO)) {
                planId = crossSchedulesPlanPO.getPlanId();
            } else {
                // 判断是星期几
                int weekday = convertWeekDay(DateUtil.dayOfWeek(todayDate));
                CrossSchedulesPlanPO weekdayPlanPO =
                        crossSchedulesPlanMapper.selectByWeekday(crossId, schedulesId, weekday);
                planId = weekdayPlanPO.getPlanId();
            }
            // 获取当前执行的方案
            List<CrossSectionPO> baseCrossSectionPOList =
                    baseCrossSectionMapper.selectByCrossAndPlan(crossId, planId);
            Integer currentSchemeId = findCurrentScheme(baseCrossSectionPOList, todayDate);
            BaseCrossSchemePO baseCrossSchemePO = baseCrossSchemeMapper.selectById(currentSchemeId);
            greenwaveCross.setOffset(baseCrossSchemePO.getOffset());
            // 构造相位时差图
            List<GreenwaveDetailVO.GreenwavePhase> greenwavePhaseList = getGreenwavePhaseList(baseCrossSchemePO);
            greenwaveCross.setGreenwavePhaseList(greenwavePhaseList);

            res.add(greenwaveCross);
        }
        return res;
    }

    private List<GreenwaveDetailVO.GreenwavePhase> getGreenwavePhaseList(BaseCrossSchemePO baseCrossSchemePO) {
        List<GreenwaveDetailVO.GreenwavePhase> res = new ArrayList<>();
        Integer schemeId = baseCrossSchemePO.getId();
        String crossId = baseCrossSchemePO.getCrossId();
        Integer offset = baseCrossSchemePO.getOffset();
        Integer coordPhaseId = baseCrossSchemePO.getCoordPhase();
        // 获取方案相位并按相位序排序
        List<CrossPhasePO> crossPhasePOList = baseCrossPhaseMapper.selectByCrossIdAndSchemeId(crossId, schemeId);
        // 减相位差，构造结果集
        int size = crossPhasePOList.size();
        for (int i = 0; i < size; i++) {
            CrossPhasePO crossPhasePO = crossPhasePOList.get(i);
            GreenwaveDetailVO.GreenwavePhase greenwavePhase = new GreenwaveDetailVO.GreenwavePhase();

            greenwavePhase.setSort(i + 1);
            greenwavePhase.setPhaseNo(crossPhasePO.getPhaseNo());
            greenwavePhase.setSeconds(crossPhasePO.getPhaseTime() - offset);
            Integer phaseId = crossPhasePO.getId();
            greenwavePhase.setIsCoordinate(ObjectUtil.equals(phaseId, coordPhaseId) ? 1 : 0);

            res.add(greenwavePhase);
        }
        // 添加最后一段相位
        GreenwaveDetailVO.GreenwavePhase greenwavePhase = new GreenwaveDetailVO.GreenwavePhase();
        CrossPhasePO crossPhasePO = crossPhasePOList.get(0);
        greenwavePhase.setSort(size + 1);
        greenwavePhase.setPhaseNo(crossPhasePO.getPhaseNo());
        greenwavePhase.setSeconds(offset);
        Integer phaseId = crossPhasePO.getId();
        greenwavePhase.setIsCoordinate(ObjectUtil.equals(phaseId, coordPhaseId) ? 1 : 0);
        res.add(greenwavePhase);
        return res;
    }

    private Integer findCurrentScheme(List<CrossSectionPO> baseCrossSectionPOList, Date todayDate) {
        Integer schemeId = null;
        for (CrossSectionPO crossSectionPO : baseCrossSectionPOList) {
            String startTimeRaw = crossSectionPO.getStartTime();
            String endTimeRaw = crossSectionPO.getEndTime();
            String startTimeStr = startTimeRaw.replace(":", "");
            String endTimeStr = endTimeRaw.replace(":", "");
            int startTime = Integer.parseInt(startTimeStr);
            int endTime = Integer.parseInt(endTimeStr);

            String todayRaw = DateUtil.format(todayDate, "HH:mm");
            String todayStr = todayRaw.replace(":", "");
            int todayTime = Integer.parseInt(todayStr);

            if (todayTime >= startTime && todayTime <= endTime) {
                schemeId = crossSectionPO.getSchemeId();
            }
        }
        return schemeId;
    }

    private int convertWeekDay(int i) {
        if (1 == i) {
            return 7;
        }
        return i - 1;
    }

    private Double calcSpeed(List<GreenwaveCrossPO> greenwaveCrossList, int i, String currentCrossId) {
        int size = greenwaveCrossList.size();
        int lastCrossIndex = size - 1;
        if (i != lastCrossIndex) { // 不是最后一个路口
            GreenwaveCrossPO nextCross = greenwaveCrossList.get(i + 1);
            String nextCrossId = nextCross.getCrossId();
            RidInfoEntity ridInfoEntity = ridInfoMapper.selectByStartEnd(currentCrossId, nextCrossId);
            Integer outDir = ridInfoEntity.getOutDir();
            CrossDirDataRealtimePO crossDirDataRealtimePO =
                    crossDirDataRealtimeMapper.selectByCrossIdAndDirType(currentCrossId, outDir);
            return crossDirDataRealtimePO.getSpeed();
        } else {
            GreenwaveCrossPO preCross = greenwaveCrossList.get(i - 1);
            String preCrossId = preCross.getCrossId();
            RidInfoEntity ridInfoEntity = ridInfoMapper.selectByStartEnd(preCrossId, currentCrossId);
            Integer inDir = ridInfoEntity.getInDir();
            CrossDirDataRealtimePO crossDirDataRealtimePO =
                    crossDirDataRealtimeMapper.selectByInDir(currentCrossId, inDir);
            return crossDirDataRealtimePO.getSpeed();
        }
    }

    private Integer calcDistance(List<GreenwaveCrossPO> greenwaveCrossList, int i, String currentCrossId) {
        GreenwaveCrossPO nextCross = greenwaveCrossList.get(i + 1);
        String nextCrossId = nextCross.getCrossId();
        RidInfoEntity ridInfoEntity = ridInfoMapper.selectByStartEnd(currentCrossId, nextCrossId);
        double length = ridInfoEntity.getLength();
        return (int)length;
    }

    @Override
    public List<GreenwaveCrossMetricsVO> greenwaveCrossMetrics(GreenwaveIdAndTimeStampBO greenwaveIdAndTimeStampBO) {
        List<GreenwaveCrossMetricsVO> res = new ArrayList<>();

        Integer greenwaveId = greenwaveIdAndTimeStampBO.getGreenwaveId();
        Long timeStamp = greenwaveIdAndTimeStampBO.getTimeStamp();
        // todo 测试用，时间戳设定为 1676082600
        timeStamp = 1676082600L;

        List<GreenwaveCrossPO> greenwaveCrossList = greenwaveCrossMapper.selectByGreenwaveId(greenwaveId);
        int crossListSize = greenwaveCrossList.size();
        int lastCrossIndex = crossListSize - 1;
        for (int i = 0; i < crossListSize; i++) {
            GreenwaveCrossPO greenwaveCrossPO = greenwaveCrossList.get(i);
            GreenwaveCrossMetricsVO greenwaveCrossMetricsVO = new GreenwaveCrossMetricsVO();
            String crossId = greenwaveCrossPO.getCrossId();
            greenwaveCrossMetricsVO.setCrossId(crossId);
            BaseCrossInfoPO baseCrossInfoPO = baseCrossInfoMapper.selectById(crossId);
            if (baseCrossInfoPO != null) {
                greenwaveCrossMetricsVO.setCrossName(baseCrossInfoPO.getName());
            }
            String currentCrossId = crossId;
            long currentSeconds = timeStamp;
            long preSeconds = currentSeconds - 60 * 5;
            // 如果不是最后一个路口
            if (i != lastCrossIndex) {
                GreenwaveCrossPO nextCross = greenwaveCrossList.get(i + 1);
                String nextCrossId = nextCross.getCrossId();
                // 计算当前路口出口方向
                RidInfoEntity ridInfoEntity = ridInfoMapper.selectByStartEnd(currentCrossId, nextCrossId);
                if (ObjectUtil.isEmpty(ridInfoEntity)) {
                    throw new RuntimeException("路网数据不连续");
                }
                Integer outDir = ridInfoEntity.getOutDir();
                List<CrossDirDataHistPO> listForNoPark =
                        crossDirDataHistMapper.selectNoPark(currentCrossId, outDir, currentSeconds, preSeconds);
                List<CrossDirDataHistPO> listForPhaseQueue =
                        crossDirDataHistMapper.selectPhaseQueue(currentCrossId, outDir, currentSeconds, preSeconds);
                setMetrics(greenwaveCrossMetricsVO, ridInfoEntity, listForNoPark, listForPhaseQueue);
            } else {
                // 最后一个路口
                GreenwaveCrossPO preCross = greenwaveCrossList.get(i - 1);
                String preCrossId = preCross.getCrossId();
                // 计算当前路口入口方向
                RidInfoEntity ridInfoEntity = ridInfoMapper.selectByStartEnd(preCrossId, currentCrossId);
                if (ObjectUtil.isEmpty(ridInfoEntity)) {
                    throw new RuntimeException("路网数据不连续");
                }
                Integer inDir = ridInfoEntity.getInDir();
                List<CrossDirDataHistPO> listForNoPark =
                        crossDirDataHistMapper.selectNoParkEnd(currentCrossId, inDir, currentSeconds, preSeconds);
                List<CrossDirDataHistPO> listForPhaseQueue =
                        crossDirDataHistMapper.selectPhaseQueueEnd(currentCrossId, inDir, currentSeconds, preSeconds);
                setMetrics(greenwaveCrossMetricsVO, ridInfoEntity, listForNoPark, listForPhaseQueue);
            }

            res.add(greenwaveCrossMetricsVO);
        }
        return res;
    }

    @Override
    @Transactional
    public void saveGreenwaveStrategy(SaveGreenwaveStrategyBO saveGreenwaveStrategyBO) {
        Integer greenwaveId = saveGreenwaveStrategyBO.getGreenwaveId();
        List<String> strategyNames = saveGreenwaveStrategyBO.getStrategyNames();
        strategyMapper.deleteGreenwave(greenwaveId);
        for (String strategyName : strategyNames) {
            // 获取策略ID
            StrategyPO strategyPO = strategyMapper.selectByName(strategyName);
            Integer strategyId = strategyPO.getId();
            // 更新绿波策略关系
            strategyMapper.insertGreenwave(greenwaveId, strategyId);
        }
    }

    private void setMetrics(GreenwaveCrossMetricsVO greenwaveCrossMetricsVO, RidInfoEntity ridInfoEntity,
                            List<CrossDirDataHistPO> listForNoPark, List<CrossDirDataHistPO> listForPhaseQueue) {
        if (CollectionUtil.isNotEmpty(listForNoPark) && CollectionUtil.isNotEmpty(listForPhaseQueue)) {
            Integer noparkPassRate = getAveRate(listForNoPark);
            greenwaveCrossMetricsVO.setNoparkPassRate(noparkPassRate);
            Integer uncoordinatePhaseQueue = getPhaseQueue(listForPhaseQueue);
            greenwaveCrossMetricsVO.setUncoordinatePhaseQueue(uncoordinatePhaseQueue);
            Double length = ridInfoEntity.getLength();
            Double speed = getSpeed(listForNoPark);
            double travelTime = length / speed;
            greenwaveCrossMetricsVO.setTrvalTime((int)travelTime);
        }
    }

    private Double getSpeed(List<CrossDirDataHistPO> listForNoPark) {
        Double total = 0.0;
        for (CrossDirDataHistPO crossDirDataHistPO : listForNoPark) {
            Double speed = crossDirDataHistPO.getSpeed();
            total += speed;
        }
        double avgSpeed = total / listForNoPark.size();
        return avgSpeed;
    }

    private Integer getPhaseQueue(List<CrossDirDataHistPO> listForPhaseQueue) {
        double maxQueue = 0.0;
        for (CrossDirDataHistPO crossDirDataHistPO : listForPhaseQueue) {
            double queueLength = crossDirDataHistPO.getQueueLength();
            if (queueLength > maxQueue) {
                maxQueue = queueLength;
            }
        }
        return (int)maxQueue;
    }

    private Integer getAveRate(List<CrossDirDataHistPO> crossDirDataHistPOList) {
        Double total = 0.0;
        for (CrossDirDataHistPO crossDirDataHistPO : crossDirDataHistPOList) {
            Double noStopRate = crossDirDataHistPO.getNoStopRate();
            total += noStopRate;
        }
        int size = crossDirDataHistPOList.size();
        double avg = total / size;
        return (int)(avg * 100);
    }

    private List<GreenWaveCrossMonitorVO> buildRes(List<GreenwaveCrossPO> greenwaveCrossPOList) {
        List<GreenWaveCrossMonitorVO> res = new ArrayList<>();
        for (GreenwaveCrossPO greenwaveCrossPO : greenwaveCrossPOList) {
            GreenWaveCrossMonitorVO greenWaveCrossMonitorVO = new GreenWaveCrossMonitorVO();

            String crossId = greenwaveCrossPO.getCrossId();
            CrossDataRealtimePO crossDataRealtimePO = crossDataRealtimeMapper.selectByCrossId(crossId);
            BaseCrossInfoPO baseCrossInfoPO = baseCrossInfoMapper.selectById(crossId);
            if (crossDataRealtimePO != null && baseCrossInfoPO != null) {
                greenWaveCrossMonitorVO.setCrossId(crossId);
                greenWaveCrossMonitorVO.setCrossName(baseCrossInfoPO.getName());
                greenWaveCrossMonitorVO.setFlow(crossDataRealtimePO.getFlow());

                List<CrossDirDataRealtimePO> crossDirDataRealtimePOList =
                        crossDirDataRealtimeMapper.selectBycrossId(crossId);
                Integer capacity = 0;
                for (CrossDirDataRealtimePO crossDirDataRealtimePO : crossDirDataRealtimePOList) {
                    capacity += crossDirDataRealtimePO.getCapacity();
                }
                greenWaveCrossMonitorVO.setCapacity(capacity);

                greenWaveCrossMonitorVO.setCongestionIndex(crossDataRealtimePO.getCongestionIndex());
                res.add(greenWaveCrossMonitorVO);
            }
        }
        return res;
    }

    private List<AbnormalCrossDetailVO.DirDataElement> buildDirData(String crossId) {
        List<AbnormalCrossDetailVO.DirDataElement> dirData = new ArrayList<>();
        // 获取当前时间之前40分钟的10位时间戳
        long currentSeconds = DateUtil.currentSeconds();
        long preSeconds = currentSeconds - 40 * 60;
        List<CrossDirDataHistPO> crossDirDataHistPOList =
                crossDirDataHistMapper.selectByCrossIdAndTimestamp(crossId, preSeconds);

        for (CrossDirDataHistPO crossDirDataHistPO : crossDirDataHistPOList) {
            AbnormalCrossDetailVO.DirDataElement dirDataElement = new AbnormalCrossDetailVO.DirDataElement();
            dirDataElement.setDir(crossDirDataHistPO.getDirType());
            // 获取采集时间
            Integer batchTime = crossDirDataHistPO.getBatchTime();
            Date date = DateUtil.date(batchTime * 1000);
            dirDataElement.setTime(date);
            // 转换状态，3、4都算拥堵
            Integer status = crossDirDataHistPO.getStatus();
            if (status == 4) {
                status = 3;
            }
            dirDataElement.setStatus(status);

            dirData.add(dirDataElement);
        }

        return dirData;
    }

    private List<AbnormalCrossDetailVO.TurnDataElement> buildTurnData(String crossId) {
        List<AbnormalCrossDetailVO.TurnDataElement> turnData = new ArrayList<>();
        List<CrossTurnDataRealtimePO> crossTurnDataRealtimePOList = crossTurnDataRealtimeMapper.selectByCrossId(crossId);
        Map<Integer, List<CrossTurnDataRealtimePO>> collect = crossTurnDataRealtimePOList.stream()
                .collect(Collectors.groupingBy(CrossTurnDataRealtimePO::getInDir));
        for (Map.Entry<Integer, List<CrossTurnDataRealtimePO>> entry : collect.entrySet()) {
            AbnormalCrossDetailVO.TurnDataElement turnDataElement = new AbnormalCrossDetailVO.TurnDataElement();
            Integer dir = entry.getKey();
            turnDataElement.setDir(dir);
            // 构造转向列表
            List<CrossTurnDataRealtimePO> subList = entry.getValue();
            List<AbnormalCrossDetailVO.TurnDataElement.TurnListElement> turnList = buildTurnList(subList);
            turnDataElement.setTurnList(turnList);
            
            turnData.add(turnDataElement);
        }
        return turnData;
    }

    private List<AbnormalCrossDetailVO.TurnDataElement.TurnListElement> buildTurnList(
            List<CrossTurnDataRealtimePO> subList) {
        List<AbnormalCrossDetailVO.TurnDataElement.TurnListElement> turnList = new ArrayList<>();

        for (CrossTurnDataRealtimePO crossTurnDataRealtimePO : subList) {
            AbnormalCrossDetailVO.TurnDataElement.TurnListElement turnListElement =
                    new AbnormalCrossDetailVO.TurnDataElement.TurnListElement();

            turnListElement.setTurn(crossTurnDataRealtimePO.getTurnType());
            turnListElement.setTurnQueueLength(crossTurnDataRealtimePO.getQueueLength());
            turnListElement.setTurnStopTimes(crossTurnDataRealtimePO.getStopTimes());
            turnListElement.setFlow(crossTurnDataRealtimePO.getFlow());

            turnList.add(turnListElement);
        }
        return turnList;
    }

    private static void buildAbnormalCrossStats(AbnormalCrossVO abnormalCrossVO,
                                                List<AbnormalCrossListVO> abnormalCrossListVOList) {
        AbnormalCrossStats abnormalCrossStats = new AbnormalCrossStats();
        Integer normal = abnormalCrossStats.getNormal();
        Integer unbalance = abnormalCrossStats.getUnbalance();
        Integer congestion = abnormalCrossStats.getCongestion();
        Integer spillover = abnormalCrossStats.getSpillover();
        for (AbnormalCrossListVO abnormalCrossListVO : abnormalCrossListVOList) {
            Integer realtimeStatus = abnormalCrossListVO.getRealtimeStatus();
            if (realtimeStatus == 0) {
                normal += 1;
            } else if (realtimeStatus == 1) {
                unbalance += 1;
            } else if (realtimeStatus == 2) {
                congestion += 1;
            } else if (realtimeStatus == 3) {
                spillover += 1;
            }
        }
        abnormalCrossStats.setNormal(normal);
        abnormalCrossStats.setUnbalance(unbalance);
        abnormalCrossStats.setCongestion(congestion);
        abnormalCrossStats.setSpillover(spillover);
        abnormalCrossVO.setAbnormalCrossStats(abnormalCrossStats);
    }

    private static void changeStatus(List<GreenwaveListVO> greenwaveListVOList) {
        if (greenwaveListVOList != null) {
            for (GreenwaveListVO greenwaveListVO : greenwaveListVOList) {
                if (greenwaveListVO.getRealtimeStatus() == 4) {
                    greenwaveListVO.setRealtimeStatus(3);
                }
            }
        }
    }

    private List<GreenwaveListVO.CrossListElement> buildCrossList(Integer greenwaveId) {
        List<GreenwaveListVO.CrossListElement> crossList = new ArrayList<>();
        List<String> crossIdList = greenwaveInfoMapper.selectCrossIdsById(greenwaveId);
        if (CollectionUtil.isNotEmpty(crossIdList)) {
            List<BaseCrossInfoPO> baseCrossInfoPOList = baseCrossInfoMapper.selectByCrossIds(crossIdList);
            for (BaseCrossInfoPO baseCrossInfoPO : baseCrossInfoPOList) {
                GreenwaveListVO.CrossListElement crossListElement = new GreenwaveListVO.CrossListElement();
                crossListElement.setCrossId(baseCrossInfoPO.getId());
                crossListElement.setName(baseCrossInfoPO.getName());
                List<Double> location = buildLocation(baseCrossInfoPO);
                crossListElement.setLocation(location);

                crossList.add(crossListElement);
            }
        }
        return crossList;
    }

    private List<Double> buildLocation(BaseCrossInfoPO baseCrossInfoPO) {
        String location = baseCrossInfoPO.getLocation();
        double[] lonLat = CrossUtil.getLonLat(location);
        List<Double> res = new ArrayList<>();
        res.add(lonLat[0]);
        res.add(lonLat[1]);
        return res;
    }

    private static void buildDuration(GreenwaveListVO greenwaveListVO) {
        Date startTime = greenwaveListVO.getStartTime();
        Date endTime = greenwaveListVO.getEndTime();
        Long duration = null;
        if (endTime != null) {
            duration = DateUtil.between(startTime, endTime, DateUnit.MINUTE);
        } else {
            duration = DateUtil.between(startTime, new Date(), DateUnit.MINUTE);
        }
        greenwaveListVO.setDuration(duration);
    }
}
